package twitter_test

import (
	"io"
	"net/http"
	"strings"
	"testing"

	"git.netflux.io/rob/elon-eats-my-tweets/generated/mocks"
	"git.netflux.io/rob/elon-eats-my-tweets/twitter"
	"github.com/stretchr/testify/assert"
)

func TestGetMe(t *testing.T) {
	testCases := []struct {
		name               string
		responseStatusCode int
		responseBody       io.Reader
		wantUser           *twitter.User
		wantErr            string
	}{
		{
			name:               "successful request",
			responseStatusCode: 200,
			responseBody:       strings.NewReader(`{"Data": {"id": "1", "name": "foo", "username": "foo bar"}}`),
			wantUser:           &twitter.User{ID: "1", Name: "foo", Username: "foo bar"},
		},
		{
			name:               "500 response",
			responseStatusCode: 500,
			responseBody:       strings.NewReader("whale"),
			wantUser:           nil,
			wantErr:            "error fetching resource: status code 500",
		},
		{
			name:               "decoder error",
			responseStatusCode: 200,
			responseBody:       strings.NewReader("<html></html>"),
			wantUser:           nil,
			wantErr:            "error decoding resource: invalid character '<' looking for beginning of value",
		},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			var getter mocks.Getter
			getter.
				On("Get", "https://api.twitter.com/2/users/me").
				Return(&http.Response{StatusCode: tc.responseStatusCode, Body: io.NopCloser(tc.responseBody)}, nil)
			client := twitter.NewAPIClient(&getter)

			user, err := client.GetMe()
			assert.Equal(t, tc.wantUser, user)

			if tc.wantErr == "" {
				assert.NoError(t, err)
			} else {
				assert.EqualError(t, err, tc.wantErr)
			}
		})
	}
}

func TestGetTweets(t *testing.T) {
	testCases := []struct {
		name               string
		userID, sinceID    string
		responseStatusCode int
		responseBody       io.Reader
		wantTweets         []*twitter.Tweet
		wantErr            string
	}{
		{
			name:               "successful request",
			userID:             "1",
			sinceID:            "1000",
			responseStatusCode: 200,
			responseBody:       strings.NewReader(`{"Data": [{"id": "101", "text": "foo"}, {"id": "102", "text": "bar"}]}`),
			wantTweets:         []*twitter.Tweet{{ID: "101", Text: "foo"}, {ID: "102", Text: "bar"}},
		},
		{
			name:               "500 response",
			userID:             "1",
			sinceID:            "1000",
			responseStatusCode: 500,
			responseBody:       strings.NewReader("whale"),
			wantTweets:         nil,
			wantErr:            "error fetching resource: status code 500",
		},
		{
			name:               "decoder error",
			userID:             "1",
			sinceID:            "1000",
			responseStatusCode: 200,
			responseBody:       strings.NewReader("<html></html>"),
			wantTweets:         nil,
			wantErr:            "error decoding resource: invalid character '<' looking for beginning of value",
		},
	}

	for _, tc := range testCases {
		t.Run(tc.name, func(t *testing.T) {
			var getter mocks.Getter
			getter.
				On("Get", "https://api.twitter.com/2/users/"+tc.userID+"/tweets?since_id="+tc.sinceID).
				Return(&http.Response{StatusCode: tc.responseStatusCode, Body: io.NopCloser(tc.responseBody)}, nil)
			client := twitter.NewAPIClient(&getter)

			tweets, err := client.GetTweets(tc.userID, tc.sinceID)
			assert.Equal(t, tc.wantTweets, tweets)

			if tc.wantErr == "" {
				assert.NoError(t, err)
			} else {
				assert.EqualError(t, err, tc.wantErr)
			}
		})
	}
}
